import re

from contextlib import suppress
from inspect import getfullargspec, isclass
from textwrap import dedent
from typing import Any, Callable, Iterable

from jinja2 import Environment, BaseLoader
from langchain.output_parsers import PydanticOutputParser
from langchain.schema import SystemMessage, HumanMessage, AIMessage

from pydantic import BaseModel


def preprocess_template(template_str):
    # Remove custom tags {% chat %} and {% endchat %}
    template_str = re.sub(r"{%\s*chat\s*%}", "", template_str)
    template_str = re.sub(r"{%\s*endchat\s*%}", "", template_str)
    # Replace {% message 'role' %} and {% endmessage %} with placeholders
    template_str = re.sub(
        r"{%\s*message\s*(\w+)\s*%}", r"<!-- message \1 -->", template_str
    )
    template_str = re.sub(r"{%\s*endmessage\s*%}", r"<!-- endmessage -->", template_str)
    return template_str


def render_chat_messages(
    context_dict: dict,
    raw_template: str,
    bypass_system: bool = False,  # Claude / o1 does not support SystemMessage
) -> list:
    environment = Environment(loader=BaseLoader())
    preprocessed_template = preprocess_template(raw_template)
    template = environment.from_string(preprocessed_template)
    rendered_template = dedent(template.render(**context_dict))

    messages = []

    # Parse the text outside chat tags and add it as a HumanMessage
    outside_chat_pattern = re.compile(r"(.*?)<!-- message", re.DOTALL)
    outside_chat_match = outside_chat_pattern.match(rendered_template)
    if outside_chat_match:
        outside_text = outside_chat_match.group(1).strip()
        if outside_text:
            messages.append(
                HumanMessage(content=[{"type": "text", "text": outside_text}])
            )

    # Parse the rendered template to construct the list of message objects
    message_pattern = re.compile(
        r"<!-- message (\w+) -->\s*(.*?)\s*<!-- endmessage -->", re.DOTALL
    )
    role_map = {
        "system": "SystemMessage" if not bypass_system else "HumanMessage",
        "ai": "AIMessage",
        "human": "HumanMessage",
    }

    for match in message_pattern.finditer(rendered_template):
        role, content = match.groups()
        message_class = role_map.get(role, "UnknownMessage")
        content = content.strip()
        if message_class == "SystemMessage":
            messages.append(SystemMessage(content=[{"type": "text", "text": content}]))
        elif message_class == "AIMessage":
            messages.append(AIMessage(content=[{"type": "text", "text": content}]))
        elif message_class == "HumanMessage":
            messages.append(HumanMessage(content=[{"type": "text", "text": content}]))
        else:
            raise ValueError(f"Unknown role: {role}")

    # Add remaining text outside chat tags as a HumanMessage
    remaining_text = message_pattern.split(rendered_template)[-1].strip()
    if remaining_text:
        messages.append(
            HumanMessage(content=[{"type": "text", "text": remaining_text}])
        )

    return messages


def format_attributes(
    func: Callable, func_args: Iterable, context_dict: dict, raw_template: str
) -> tuple[str, dict]:
    """
    Formats the attributes of the method's parent class / obj.
    """
    varnames = func.__code__.co_varnames
    if "self" in varnames or "cls" in varnames:
        _reference = "self" if "self" in varnames else "cls"
        _object = func_args[0]
        placeholders = re.findall(r"\{\{(.*?)\}\}|\{%\s*(.*?)\s*%\}", raw_template)
        placeholders = [
            item.strip()
            for sublist in placeholders
            for item in sublist
            if item and _reference in item
        ]
        placeholders = list(set(placeholders))
        for placeholder in placeholders:
            _attribute = placeholder.split(f"{_reference}.")[-1]
            if "." in _attribute:
                attr_template_name = _attribute.replace(".", "__")
            else:
                attr_template_name = _attribute
            context_dict[attr_template_name] = get_nested_attr(_object, _attribute)
            raw_template = raw_template.replace(
                f"{_reference}.{_attribute}", attr_template_name
            )
    return raw_template, context_dict


def format_args(func: Callable, func_args: Iterable, context_dict: dict) -> dict:
    """
    Formats the arguments of a function.
    """
    if len(func_args) > 0:
        args_names = getfullargspec(func).args
        self_is_here = "self" in args_names
        if self_is_here:
            args_names = args_names[1:]
        for i, arg_name in enumerate(args_names):
            with suppress(IndexError):
                context_dict[arg_name] = func_args[i + 1 if self_is_here else i]
    return context_dict


def format_func_body_result(
    result: Any, raw_template: str, context_dict: dict
) -> tuple[str, dict]:
    """
    Formats the result of a function to be used in a prompt.
    """
    if result and "__result__" in raw_template:
        context_dict["result_"] = result
        raw_template = raw_template.replace("__result__", "result_")
    return raw_template, context_dict


def format_return_type_instructions(return_type: Any, raw_template: str) -> str:
    """
    Returns the format instructions of a Pydantic schema.
    """
    if return_type and isclass(return_type) and issubclass(return_type, BaseModel):
        raw_template += "\n\n" + output_parser(return_type).get_format_instructions()
    elif (
        return_type
        and not isclass(return_type)
        and return_type != Any
        and return_type != None
        and return_type != str
    ):
        raw_template += (
            "\n\n"
            "Return the result as a string representing a perfectly exact python expression "
            f"respecting this type annotation: {return_type}\n"
            "Answer ONLY by the python expression, without any additional text. "
            "Do not enclose your answer in markdown code blocks."
        )
    return raw_template


def get_nested_attr(obj: Any, attr: str) -> Any:
    """
    Recusrively gets a nested attribute from an object.
    """
    for attr in attr.split("."):
        if hasattr(obj, attr) and not callable(getattr(obj, attr)):
            obj = getattr(obj, attr)
    return obj


def output_parser(schema: BaseModel) -> PydanticOutputParser:
    """
    Converts a Pydantic scheme to a LCEL compatible output parser.
    Can be used to either get format instructions in a prompt or to parse the llm output.
    """
    return PydanticOutputParser(pydantic_object=schema)


def clean_docstring(docstring: str) -> str:
    """
    Cleans a docstring from unnecessary spaces and new lines.
    """
    docstring = docstring.strip()
    docstring = re.sub(" +", " ", docstring)
    docstring = re.sub("\n+", "\n", docstring)
    return docstring



import os
import json

from copy import deepcopy
from functools import wraps
from inspect import signature, isclass
from pprint import pprint
from typing import Any, Callable

from dotenv import load_dotenv
from langchain_core.runnables import RunnableLambda, RunnableParallel
from langchain_core.language_models.chat_models import BaseChatModel
from langchain.output_parsers import RetryWithErrorOutputParser
from langchain.prompts import ChatPromptTemplate
from langchain.schema import SystemMessage, HumanMessage
from langchain_openai import ChatOpenAI
from langchain_anthropic import ChatAnthropic
from .images import convert_image_to_base64
from pydantic import BaseModel
from .utils import (
    clean_docstring,
    format_args,
    format_attributes,
    format_func_body_result,
    format_return_type_instructions,
    output_parser,
    render_chat_messages,
)


load_dotenv()


def ask(
    call: bool = True,
    verbose: bool = False,
    return_prompt_only: bool = False,
    return_as_dict: bool = False,
    model_name: str = "gpt-4o",
    max_tokens: int = 4096,
    image_quality: str = "high",
    chat_model_class: BaseChatModel = ChatOpenAI,
    **llm_kwargs,
):
    def decorator(func: Callable) -> Callable:

        @wraps(func)
        def wrapper(*args, **kwargs) -> Callable:
            api_key_from_kwargs = kwargs.pop("api_key", None)
            context_dict = deepcopy(kwargs)
            image = context_dict.get("image", None)
            return_type = signature(func).return_annotation
            docstring = func.__doc__
            messages = []

            is_claude = model_name.startswith("claude")
            is_o1 = model_name.startswith("o1")
            bypass_system_msg = is_claude or is_o1

            if is_claude:
                chat_model_class = ChatAnthropic
                api_key_from_llm_kwargs = llm_kwargs.get("anthropic_api_key", None)
                llm_kwargs["anthropic_api_key"] = (
                    api_key_from_kwargs
                    or api_key_from_llm_kwargs
                    or os.getenv("ANTHROPIC_API_KEY")
                )

            elif model_name.startswith("gpt") or is_o1:
                chat_model_class = ChatOpenAI
                api_key_from_llm_kwargs = llm_kwargs.get("openai_api_key", None)
                llm_kwargs["openai_api_key"] = (
                    api_key_from_kwargs
                    or api_key_from_llm_kwargs
                    or os.getenv("OPENAI_API_KEY")
                )

            if verbose:
                print("RETURN TYPE: ", return_type)

            if docstring:
                raw_template = clean_docstring(docstring)
                raw_template, context_dict = format_func_body_result(
                    result=func(*args, **kwargs),
                    raw_template=raw_template,
                    context_dict=context_dict,
                )

                raw_template = format_return_type_instructions(
                    return_type=return_type, raw_template=raw_template
                )

                raw_template, context_dict = format_attributes(
                    func=func,
                    func_args=args,
                    context_dict=context_dict,
                    raw_template=raw_template,
                )

                context_dict = format_args(
                    func=func, func_args=args, context_dict=context_dict
                )
                messages.extend(
                    render_chat_messages(
                        context_dict,
                        raw_template,
                        bypass_system=bypass_system_msg,
                    )
                )

            else:
                msg_class = SystemMessage if not bypass_system_msg else HumanMessage
                messages.append(
                    msg_class(content=[{"type": "text", "text": " ".join(args)}])
                )

            if image:
                image_base64 = convert_image_to_base64(image)
                messages.append(
                    HumanMessage(
                        content=[
                            {
                                "type": "image_url",
                                "image_url": {
                                    "url": image_base64,
                                    "detail": image_quality,
                                },
                            }
                        ]
                    )
                )

            if verbose:
                print("FORMATTED PROMPT MESSAGES:")
                for message in messages:
                    pprint(message.dict())

            prompt = ChatPromptTemplate.from_messages(messages)
            if return_prompt_only:
                return prompt

            llm_model_class_kwargs = {
                "model_name": model_name,
                "max_tokens": max_tokens,
                **llm_kwargs,
            }
            if is_o1:
                llm_model_class_kwargs.pop("max_tokens")
                llm_model_class_kwargs["temperature"] = 1

            llm = chat_model_class(**llm_model_class_kwargs)
            if (
                return_type
                and isclass(return_type)
                and issubclass(return_type, BaseModel)
            ):
                parser = output_parser(return_type)
                retry_parser = RetryWithErrorOutputParser.from_llm(
                    llm=llm, parser=parser
                )
                completion_chain = prompt | llm
                main_chain = (
                    RunnableParallel(
                        completion=completion_chain,
                        prompt_value=prompt,
                    )
                    | RunnableLambda(
                        lambda x: retry_parser.parse_with_prompt(
                            completion=x.get("completion").content,
                            prompt_value=x.get("prompt_value"),
                        )
                    )
                    | RunnableLambda(
                        lambda x: json.loads(x.json()) if return_as_dict else x
                    )
                )
                return main_chain.invoke({}) if call else main_chain

            elif (
                return_type
                and not isclass(return_type)
                and return_type != str
                and return_type != Any
                and return_type != None
            ):
                return eval((prompt | llm).invoke({}).content) if call else prompt | llm

            else:
                return (prompt | llm).invoke({}).content if call else prompt | llm

        return wrapper

    return decorator



from setuptools import setup, find_packages

setup(
    name="ask-llm",
    version="0.1.183",
    packages=find_packages(),
    url="https://github.com/FlorianMgs/ask-llm",
    license="MIT",
    author="Florian Magisson",
    author_email="madgic.shrooms@gmail.com",
    description="The easiest way to supercharge your apps with LLM!",
    install_requires=[
        "Jinja2>=3.1.3",
        "pillow>=10.2.0",
        "langchain>=0.3.2",
        "langchain-community>=0.3.1",
        "langchain-core>=0.3.9",
        "langchain-openai>=0.2.0",
        "langchain-anthropic>=0.2.3",
        "python-dotenv>=1.0.1",
    ],
)



import base64
import requests
from PIL import Image
from io import BytesIO

import os


def is_base64(sb):
    try:
        if isinstance(sb, str):
            base64.b64decode(sb)
            return True
        else:
            return False
    except Exception:
        return False


def is_url(s):
    return s.startswith("http://") or s.startswith("https://")


def is_filepath(s):
    return os.path.isfile(s)


def image_to_b64(image: Image.Image, image_format: str | None = None) -> str:
    image_io = BytesIO()
    image_format = image_format.upper() if image_format else "WEBP"
    try:
        image.save(image_io, image_format)
    except KeyError:
        image.save(image_io, format="WEBP")
    encoded = base64.b64encode(image_io.getvalue()).decode()
    return f"data:image/{image_format.lower()};base64,{encoded}"


def download_image_as_pil_image(url: str) -> Image.Image:
    response = requests.get(url)
    return Image.open(BytesIO(response.content))


def convert_image_to_base64(input_image: str | Image.Image) -> str:
    if isinstance(input_image, Image.Image):
        return image_to_b64(input_image)
    elif isinstance(input_image, str):
        if is_base64(input_image):
            return input_image
        elif is_url(input_image):
            pil_image = download_image_as_pil_image(input_image)
            b64 = image_to_b64(pil_image)
            return b64
        elif is_filepath(input_image):
            pil_image = Image.open(input_image)
            return image_to_b64(pil_image)
        else:
            raise ValueError(
                "Invalid input: Input string is not a valid path, URL, or base64 string."
            )
    else:
        raise TypeError(
            "Invalid input type: Input must be a PIL Image, a file path, a URL, or a base64 string."
        )
